@@ -4,13 +4,20 @@ from __future__ import division
 import numpy as np
 import os
 import random
-import concurrent.futures
+import logging
+from multiprocessing import Pool
 
 try:
     from sorn.utils import Initializer
 except:
     from utils import Initializer
 
+logging.basicConfig(
+    level=logging.INFO,
+    format="%(asctime)s:%(levelname)s:%(message)s",
+    handlers=[logging.FileHandler("sorn.log"), logging.StreamHandler()],
+)
+
 
 class Sorn(object):
 
@@ -140,7 +147,10 @@ class Plasticity(Sorn):
         self.te_min = Sorn.te_min  # Excitatory minimum Threshold
         self.te_max = Sorn.te_max  # Excitatory maximum Threshold
 
-    def stdp(self, wee: np.array, x: np.array, cutoff_weights: list):
+    def stdp(
+        self,
+        params,
+    ):
         """Apply STDP rule : Regulates synaptic strength between the pre(Xj) and post(Xi) synaptic neurons
 
         Args:
@@ -153,38 +163,44 @@ class Plasticity(Sorn):
         Returns:
             wee (array):  Weight matrix
         """
+        wee, x, cutoff_weights, freeze = params[0], params[1], params[2], params[3]
 
-        x = np.asarray(x)
-        xt_1 = x[:, 0]
-        xt = x[:, 1]
-        wee_t = wee.copy()
+        if freeze:
+            return wee
 
-        # STDP applies only on the neurons which are connected.
+        else:
+            x = np.asarray(x)
+            xt_1 = x[:, 0]
+            xt = x[:, 1]
+            wee_t = wee.copy()
 
-        for i in range(len(wee_t[0])):  # Each neuron i, Post-synaptic neuron
+            # STDP applies only on the neurons which are connected.
 
-            for j in range(
-                len(wee_t[0:])
-            ):  # Incoming connection from jth pre-synaptic neuron to ith neuron
+            for i in range(len(wee_t[0])):  # Each neuron i, Post-synaptic neuron
 
-                if wee_t[j][i] != 0.0:  # Check connectivity
+                for j in range(
+                    len(wee_t[0:])
+                ):  # Incoming connection from jth pre-synaptic neuron to ith neuron
 
-                    # Get the change in weight
-                    delta_wee_t = self.eta_stdp * (xt[i] * xt_1[j] - xt_1[i] * xt[j])
+                    if wee_t[j][i] != 0.0:  # Check connectivity
 
-                    # Update the weight between jth neuron to i ""Different from notation in article
+                        # Get the change in weight
+                        delta_wee_t = self.eta_stdp * (
+                            xt[i] * xt_1[j] - xt_1[i] * xt[j]
+                        )
 
-                    wee_t[j][i] = wee[j][i] + delta_wee_t
+                        # Update the weight between jth neuron to i ""Different from notation in article
 
-        # Prune the smallest weights induced by plasticity mechanisms; Apply lower cutoff weight
-        wee_t = Initializer.prune_small_weights(wee_t, cutoff_weights[0])
+                        wee_t[j][i] = wee[j][i] + delta_wee_t
 
-        # Check and set all weights < upper cutoff weight
-        wee_t = Initializer.set_max_cutoff_weight(wee_t, cutoff_weights[1])
+            # Prune the smallest weights induced by plasticity mechanisms; Apply lower cutoff weight
+            wee = Initializer.prune_small_weights(wee_t, cutoff_weights[0])
 
-        return wee_t
+            # Check and set all weights < upper cutoff weight
+            wee = Initializer.set_max_cutoff_weight(wee, cutoff_weights[1])
+            return wee
 
-    def ip(self, te: np.array, x: np.array):
+    def ip(self, params):
         """Intrinsic Plasiticity mechanism
 
         Args:
@@ -196,35 +212,29 @@ class Plasticity(Sorn):
             te (array): Threshold vector of excitatory units
         """
 
-        # IP rule: Active unit increases its threshold and inactive decreases its threshold.
-        xt = x[:, 1]
-
-        te_update = te + self.eta_ip * (xt.reshape(self.ne, 1) - self.h_ip)
-
-        # Check whether all te are in range [0.0,1.0] and update acordingly
-
-        # Update te < 0.0 ---> 0.0
-        # te_update = prune_small_weights(te_update,self.te_min)
-
-        # Set all te > 1.0 --> 1.0
-        # te_update = set_max_cutoff_weight(te_update,self.te_max)
-
-        return te_update
+        te, x, freeze = (params[0], params[1], params[2])
+        if freeze:
+            return te
+        else:
+            # IP rule: Active unit increases its threshold and inactive decreases its threshold.
+            xt = x[:, 1]
+            te = te + self.eta_ip * (xt.reshape(self.ne, 1) - self.h_ip)
+            return te
 
     @staticmethod
-    def ss(wee: np.array):
+    def ss(param):
         """Synaptic Scaling or Synaptic Normalization
 
         Args:
-            wee (array):  Weight matrix
+            param (array):  Weight matrix
 
         Returns:
-            wee (array):  Scaled Weight matrix
+            param (array):  Scaled Weight matrix
         """
-        wee = wee / np.sum(wee, axis=0)
-        return wee
+        param = param / np.sum(param, axis=0)
+        return param
 
-    def istdp(self, wei: np.array, x: np.array, y: np.array, cutoff_weights: list):
+    def istdp(self, params):
         """Apply iSTDP rule, which regulates synaptic strength between the pre inhibitory(Xj) and post Excitatory(Xi) synaptic neurons
 
         Args:
@@ -236,54 +246,68 @@ class Plasticity(Sorn):
 
             cutoff_weights (list): Maximum and minimum weight ranges
 
+            freeze (bool): If True, skip the structural plasticity step. Default False
+
         Returns:
             wei (array): Synaptic strengths from inhibitory to excitatory"""
 
-        # Excitatory network activity
-        xt = np.asarray(x)[:, 1]
+        wei, x, y, cutoff_weights, freeze = (
+            params[0],
+            params[1],
+            params[2],
+            params[3],
+            params[4],
+        )
 
-        # Inhibitory network activity
-        yt_1 = np.asarray(y)[:, 0]
+        if freeze:
+            return wei
+        else:
+            # Excitatory network activity
+            xt = np.asarray(x)[:, 1]
 
-        # iSTDP applies only on the neurons which are connected.
-        wei_t = wei.copy()
+            # Inhibitory network activity
+            yt_1 = np.asarray(y)[:, 0]
 
-        for i in range(
-            len(wei_t[0])
-        ):  # Each neuron i, Post-synaptic neuron: means for each column;
+            # iSTDP applies only on the neurons which are connected.
+            wei_t = wei.copy()
 
-            for j in range(
-                len(wei_t[0:])
-            ):  # Incoming connection from j, pre-synaptic neuron to ith neuron
+            for i in range(
+                len(wei_t[0])
+            ):  # Each neuron i, Post-synaptic neuron: means for each column;
 
-                if wei_t[j][i] != 0.0:  # Check connectivity
+                for j in range(
+                    len(wei_t[0:])
+                ):  # Incoming connection from j, pre-synaptic neuron to ith neuron
 
-                    # Get the change in weight
-                    delta_wei_t = (
-                        -self.eta_inhib * yt_1[j] * (1 - xt[i] * (1 + 1 / self.mu_ip))
-                    )
+                    if wei_t[j][i] != 0.0:  # Check connectivity
 
-                    # Update the weight between jth neuron to i ""Different from notation in article
+                        # Get the change in weight
+                        delta_wei_t = (
+                            -self.eta_inhib
+                            * yt_1[j]
+                            * (1 - xt[i] * (1 + 1 / self.mu_ip))
+                        )
 
-                    wei_t[j][i] = wei[j][i] + delta_wei_t
+                        # Update the weight between jth neuron to i ""Different from notation in article
 
-        # Prune the smallest weights induced by plasticity mechanisms; Apply lower cutoff weight
-        wei_t = Initializer.prune_small_weights(wei_t, cutoff_weights[0])
+                        wei_t[j][i] = wei[j][i] + delta_wei_t
 
-        # Check and set all weights < upper cutoff weight
-        wei_t = Initializer.set_max_cutoff_weight(wei_t, cutoff_weights[1])
+            # Prune the smallest weights induced by plasticity mechanisms; Apply lower cutoff weight
+            wei = Initializer.prune_small_weights(wei_t, cutoff_weights[0])
 
-        return wei_t
+            # Check and set all weights < upper cutoff weight
+            wei = Initializer.set_max_cutoff_weight(wei, cutoff_weights[1])
+            return wei
 
     @staticmethod
-    def structural_plasticity(wee: np.array):
+    def structural_plasticity(param):
         """Add new connection value to the smallest weight between excitatory units randomly
 
         Args:
-            wee (array): Weight matrix
+            param (array): Weight matrix
 
         Returns:
-            wee (array):  Weight matrix"""
+            param (array):  Weight matrix"""
 
         p_c = np.random.randint(0, 10, 1)
 
@@ -291,7 +315,7 @@ class Plasticity(Sorn):
 
             # Do structural plasticity
             # Choose the smallest weights randomly from the weight matrix wee
-            indexes = Initializer.get_unconnected_indexes(wee)
+            indexes = Initializer.get_unconnected_indexes(param)
 
             # Choose any idx randomly such that i!=j
             while True:
@@ -299,9 +323,9 @@ class Plasticity(Sorn):
                 if idx_rand[0] != idx_rand[1]:
                     break
 
-            wee[idx_rand[0]][idx_rand[1]] = 0.001
+            param[idx_rand[0]][idx_rand[1]] = 0.001
 
-        return wee
+        return param
 
     @staticmethod
     def initialize_plasticity():
@@ -334,7 +358,6 @@ class Plasticity(Sorn):
         )
 
         Wee_init = Initializer.zero_sum_incoming_check(WEE_init)
-        # Wei_init = initializer.zero_sum_incoming_check(WEI_init.T)  # For SORN 1
         Wei_init = Initializer.zero_sum_incoming_check(WEI_init)
         Wie_init = Initializer.zero_sum_incoming_check(WIE_init)
 
@@ -342,9 +365,9 @@ class Plasticity(Sorn):
         v = np.count_nonzero(Wei_init)
         b = np.count_nonzero(Wie_init)
 
-        print("Network Initialized")
-        print("Number of connections in Wee %s , Wei %s, Wie %s" % (c, v, b))
-        print(
+        logging.info("Network Initialized")
+        logging.info("Number of connections in Wee %s , Wei %s, Wie %s" % (c, v, b))
+        logging.info(
             "Shapes Wee %s Wei %s Wie %s"
             % (Wee_init.shape, Wei_init.shape, Wie_init.shape)
         )
@@ -373,54 +396,13 @@ class Plasticity(Sorn):
         return wee, wei, wie, te, ti, x, y
 
 
-class Async:
-    def __init__(self, max_workers=4):
-        super().__init__()
-        self.max_workers = max_workers
-        self.plasticity = Plasticity()
-
-    def step(self, X, Y, Wee, Wei, Te, freeze):
-
-        with concurrent.futures.ThreadPoolExecutor(
-            max_workers=self.max_workers
-        ) as executor:
-
-            # STDP
-            if "stdp" not in freeze:
-                stdp = executor.submit(
-                    self.plasticity.stdp, Wee, X, cutoff_weights=(0.0, 1.0)
-                )
-                Wee = stdp.result()
-
-            # Intrinsic plasticity
-            if "ip" not in freeze:
-                ip = executor.submit(self.plasticity.ip, Te, X)
-                Te = ip.result()
-            # Structural plasticity
-            if "sp" not in freeze:
-                sp = executor.submit(self.plasticity.structural_plasticity, Wee)
-                Wee = sp.result()
-            # iSTDP
-            if "istdp" not in freeze:
-                istdp = executor.submit(
-                    self.plasticity.istdp, Wei, X, Y, cutoff_weights=(0.0, 1.0)
-                )
-                Wei = istdp.result()
-
-        # Synaptic scaling Wee
-        if "ss" not in freeze:
-            Wee = self.plasticity.ss(Wee)
-            Wei = self.plasticity.ss(Wei)
-        return Wee, Wei, Te
-
-
 class MatrixCollection(Sorn):
-    """Collect all matrices initialized and updated during simulation (plasiticity and training phases)
+    """Collect all matrices initialized and updated during simulation(plasiticity and training phases)
 
     Args:
-        phase (str): Training or Plasticity phase
+        phase(str): Training or Plasticity phase
 
-        matrices (dict): Network activity, threshold and connection matrices
+        matrices(dict): Network activity, threshold and connection matrices
 
     Returns:
         MatrixCollection instance"""
@@ -500,15 +482,15 @@ class MatrixCollection(Sorn):
         """Update weight matrices
 
         Args:
-            wee (array): Excitatory-Excitatory weight matrix
+            wee(array): Excitatory-Excitatory weight matrix
 
-            wei (array): Inhibitory-Excitatory weight matrix
+            wei(array): Inhibitory-Excitatory weight matrix
 
-            wie (array): Excitatory-Inhibitory weight matrix
+            wie(array): Excitatory-Inhibitory weight matrix
 
-            i (int): Time step
+            i(int): Time step
         Returns:
-            tuple (array): Weight Matrices Wee, Wei, Wie"""
+            tuple(array): Weight Matrices Wee, Wei, Wie"""
 
         self.Wee[i + 1] = wee
         self.Wei[i + 1] = wei
@@ -520,14 +502,14 @@ class MatrixCollection(Sorn):
         """Update threshold matrices
 
         Args:
-            te (array): Excitatory threshold
+            te(array): Excitatory threshold
 
-            ti (array): Inhibitory threshold
+            ti(array): Inhibitory threshold
 
-            i (int): Time step
+            i(int): Time step
 
         Returns:
-            tuple (array): Threshold Matrices Te and Ti"""
+            tuple(array): Threshold Matrices Te and Ti"""
 
         self.Te[i + 1] = te
         self.Ti[i + 1] = ti
@@ -539,14 +521,14 @@ class MatrixCollection(Sorn):
         """Network state at current time step
 
         Args:
-            excitatory_net (array): Excitatory network activity
+            excitatory_net(array): Excitatory network activity
 
-            inhibitory_net (array): Inhibitory network activity
+            inhibitory_net(array): Inhibitory network activity
 
-            i (int): Time step
+            i(int): Time step
 
         Returns:
-            tuple (array): Updated Excitatory and Inhibitory states
+            tuple(array): Updated Excitatory and Inhibitory states
         """
 
         self.X[i + 1] = excitatory_net
@@ -558,14 +540,14 @@ class MatrixCollection(Sorn):
         """Network activity at previous time step
 
         Args:
-            x (array): Excitatory network activity
+            x(array): Excitatory network activity
 
-            y (array): Inhibitory network activity
+            y(array): Inhibitory network activity
 
-            i (int): Time step
+            i(int): Time step
 
         Returns:
-            tuple (array): Previous Excitatory and Inhibitory states
+            tuple(array): Previous Excitatory and Inhibitory states
         """
         x_1, y_1 = [0] * self.time_steps, [0] * self.time_steps
         x_1[i] = x
@@ -579,10 +561,10 @@ class NetworkState(Plasticity):
     """The evolution of network states
 
     Args:
-        v_t (array): External input/stimuli
+        v_t(array): External input/stimuli
 
     Returns:
-        instance (object): NetworkState instance"""
+        instance(object): NetworkState instance"""
 
     def __init__(self, v_t: np.array):
         super().__init__()
@@ -600,12 +582,12 @@ class NetworkState(Plasticity):
         """Excitatory Post synaptic potential towards neurons in the reservoir in the absence of external input
 
         Args:
-            weights (array): Synaptic strengths
+            weights(array): Synaptic strengths
 
-            activity_vector (list): Acitivity of inhibitory or Excitatory neurons
+            activity_vector(list): Acitivity of inhibitory or Excitatory neurons
 
         Returns:
-            incoming (array): Excitatory Post synaptic potential towards neurons
+            incoming(array): Excitatory Post synaptic potential towards neurons
         """
         incoming = weights * activity_vector
         incoming = np.array(incoming.sum(axis=0))
@@ -623,20 +605,20 @@ class NetworkState(Plasticity):
         """Activity of Excitatory neurons in the network
 
         Args:
-            wee (array): Excitatory-Excitatory weight matrix
+            wee(array): Excitatory-Excitatory weight matrix
 
-            wei (array): Inhibitory-Excitatory weight matrix
+            wei(array): Inhibitory-Excitatory weight matrix
 
-            te (array): Excitatory threshold
+            te(array): Excitatory threshold
 
-            x (array): Excitatory network activity
+            x(array): Excitatory network activity
 
-            y (array): Inhibitory network activity
+            y(array): Inhibitory network activity
 
-            white_noise_e (array): Gaussian noise
+            white_noise_e(array): Gaussian noise
 
         Returns:
-            x (array): Current Excitatory network activity
+            x(array): Current Excitatory network activity
         """
         xt = x[:, 1]
         xt = xt.reshape(self.ne, 1)
@@ -668,18 +650,18 @@ class NetworkState(Plasticity):
         """Activity of Excitatory neurons in the network
 
         Args:
-            wee (array): Excitatory-Excitatory weight matrix
+            wee(array): Excitatory-Excitatory weight matrix
 
-            wie (array): Excitatory-Inhibitory weight matrix
+            wie(array): Excitatory-Inhibitory weight matrix
 
-            ti (array): Inhibitory threshold
+            ti(array): Inhibitory threshold
 
-            y (array): Inhibitory network activity
+            y(array): Inhibitory network activity
 
-            white_noise_i (array): Gaussian noise
+            white_noise_i(array): Gaussian noise
 
         Returns:
-            y (array): Current Inhibitory network activity"""
+            y(array): Current Inhibitory network activity"""
 
         wie = np.asarray(wie)
         yt = y[:, 1]
@@ -708,20 +690,20 @@ class NetworkState(Plasticity):
 
         Args:
 
-            wee (array): Excitatory-Excitatory weight matrix
+            wee(array): Excitatory-Excitatory weight matrix
 
-            wei (array): Inhibitory-Excitatory weight matrix
+            wei(array): Inhibitory-Excitatory weight matrix
 
-            te (array): Excitatory threshold
+            te(array): Excitatory threshold
 
-            x (array): Excitatory network activity
+            x(array): Excitatory network activity
 
-            y (array): Inhibitory network activity
+            y(array): Inhibitory network activity
 
-            white_noise_e (array): Gaussian noise
+            white_noise_e(array): Gaussian noise
 
         Returns:
-            xt (array): Recurrent network state
+            xt(array): Recurrent network state
         """
         xt = x[:, 1]
         xt = xt.reshape(self.ne, 1)
@@ -749,26 +731,26 @@ class Simulator_(Sorn):
     """Simulate SORN using external input/noise using the fresh or pretrained matrices
 
     Args:
-        inputs (np.array, optional): External stimuli. Defaults to None.
+        inputs(np.array, optional): External stimuli. Defaults to None.
 
-        phase (str, optional): Plasticity phase. Defaults to "plasticity".
+        phase(str, optional): Plasticity phase. Defaults to "plasticity".
 
-        matrices (dict, optional): Network states, connections and threshold matrices. Defaults to None.
+        matrices(dict, optional): Network states, connections and threshold matrices. Defaults to None.
 
-        time_steps (int, optional): Total number of time steps to simulate the network. Defaults to 1.
+        time_steps(int, optional): Total number of time steps to simulate the network. Defaults to 1.
 
-        noise (bool, optional): If True, noise will be added. Defaults to True.
+        noise(bool, optional): If True, noise will be added. Defaults to True.
 
     Returns:
-        plastic_matrices (dict): Network states, connections and threshold matrices
+        plastic_matrices(dict): Network states, connections and threshold matrices
 
-        X_all (array): Excitatory network activity collected during entire simulation steps
+        X_all(array): Excitatory network activity collected during entire simulation steps
 
-        Y_all (array): Inhibitory network activity collected during entire simulation steps
+        Y_all(array): Inhibitory network activity collected during entire simulation steps
 
-        R_all (array): Recurrent network activity collected during entire simulation steps
+        R_all(array): Recurrent network activity collected during entire simulation steps
 
-        frac_pos_active_conn (list): Number of positive connection strengths in the network at each time step during simulation"""
+        frac_pos_active_conn(list): Number of positive connection strengths in the network at each time step during simulation"""
 
     def __init__(self):
         super().__init__()
@@ -782,36 +764,33 @@ class Simulator_(Sorn):
         time_steps: int = None,
         noise: bool = True,
         freeze: list = None,
-        max_workers: int = 4,
         **kwargs
     ):
         """Simulation/Plasticity phase
 
         Args:
-            inputs (np.array, optional): External stimuli. Defaults to None.
+            inputs(np.array, optional): External stimuli. Defaults to None.
 
-            phase (str, optional): Plasticity phase. Defaults to "plasticity"
+            phase(str, optional): Plasticity phase. Defaults to "plasticity"
 
-            matrices (dict, optional): Network states, connections and threshold matrices. Defaults to None.
+            matrices(dict, optional): Network states, connections and threshold matrices. Defaults to None.
 
-            time_steps (int, optional): Total number of time steps to simulate the network. Defaults to 1.
+            time_steps(int, optional): Total number of time steps to simulate the network. Defaults to 1.
 
-            noise (bool, optional): If True, noise will be added. Defaults to True.
+            noise(bool, optional): If True, noise will be added. Defaults to True.
 
-            freeze (list, optional): List of synaptic plasticity mechanisms which will be turned off during simulation. Defaults to None.
-
-            max_workers (int, optional): Maximum workers for multhreading the plasticity steps
+            freeze(list, optional): List of synaptic plasticity mechanisms which will be turned off during simulation. Defaults to None.
 
         Returns:
-            plastic_matrices (dict): Network states, connections and threshold matrices
+            plastic_matrices(dict): Network states, connections and threshold matrices
 
-            X_all (array): Excitatory network activity collected during entire simulation steps
+            X_all(array): Excitatory network activity collected during entire simulation steps
 
-            Y_all (array): Inhibitory network activity collected during entire simulation steps
+            Y_all(array): Inhibitory network activity collected during entire simulation steps
 
-            R_all (array): Recurrent network activity collected during entire simulation steps
+            R_all(array): Recurrent network activity collected during entire simulation steps
 
-            frac_pos_active_conn (list): Number of positive connection strengths in the network at each time step during simulation"""
+            frac_pos_active_conn(list): Number of positive connection strengths in the network at each time step during simulation"""
 
         assert (
             phase == "plasticity" or "training"
@@ -822,7 +801,6 @@ class Simulator_(Sorn):
         self.phase = phase
         self.matrices = matrices
         self.freeze = [] if freeze == None else freeze
-        self.max_workers = max_workers
         kwargs_ = [
             "ne",
             "nu",
@@ -848,6 +826,7 @@ class Simulator_(Sorn):
         # assert Sorn.nu == len(inputs[:,0]),"Size mismatch: Input != Nu "
         Sorn.ni = int(0.2 * Sorn.ne)
 
+        self.plasticity = Plasticity()
         # Initialize/Get the weight, threshold matrices and activity vectors
         matrix_collection = MatrixCollection(phase=self.phase, matrices=self.matrices)
 
@@ -856,7 +835,8 @@ class Simulator_(Sorn):
         X_all = [0] * self.time_steps
         Y_all = [0] * self.time_steps
         R_all = [0] * self.time_steps
-
+        Te_all = [0] * self.time_steps
+        Wee_all = [0] * self.time_steps
         frac_pos_active_conn = []
 
         # To get the last activation status of Exc and Inh neurons
@@ -911,9 +891,32 @@ class Simulator_(Sorn):
             y_buffer[:, 0] = Y[i][:, 1]
             y_buffer[:, 1] = inhibitory_state_yt_buffer.T
 
-            Wee[i], Wei[i], Te[i] = Async(max_workers=max_workers).step(
-                x_buffer, y_buffer, Wee[i], Wei[i], Te[i], self.freeze
+            # Plasticity phase
+            plasticity = Plasticity()
+            pool = Pool(processes=4)
+
+            stdp = pool.apply_async(
+                plasticity.stdp, [(Wee[i], x_buffer, (0.0, 1.0), "stdp" in self.freeze)]
             )
+
+            ip = pool.apply_async(
+                plasticity.ip, [(Te[i], x_buffer, "ip" in self.freeze)]
+            )
+
+            istdp = pool.apply_async(
+                plasticity.istdp,
+                [(Wei[i], x_buffer, y_buffer, (0.0, 1.0), "istdp" in self.freeze)],
+            )
+            Wee[i], Te[i], Wei[i] = stdp.get(), ip.get(), istdp.get()
+
+            if "sp" not in self.freeze:
+                Wee[i] = plasticity.structural_plasticity(Wee[i])
+
+            if "ss" not in self.freeze:
+
+                Wee[i] = plasticity.ss(Wee[i])
+                Wei[i] = plasticity.ss(Wei[i])
+
             # Assign the matrices to the matrix collections
             matrix_collection.weight_matrix(Wee[i], Wei[i], Wie[i], i)
             matrix_collection.threshold_matrix(Te[i], Ti[i], i)
@@ -922,6 +925,8 @@ class Simulator_(Sorn):
             X_all[i] = x_buffer[:, 1]
             Y_all[i] = y_buffer[:, 1]
             R_all[i] = r
+            Te_all[i] = Te[i]
+            Wee_all[i] = Wee[i]
 
         plastic_matrices = {
             "Wee": matrix_collection.Wee[-1],
@@ -933,7 +938,15 @@ class Simulator_(Sorn):
             "Y": Y[-1],
         }
 
-        return plastic_matrices, X_all, Y_all, R_all, frac_pos_active_conn
+        return (
+            plastic_matrices,
+            X_all,
+            Y_all,
+            R_all,
+            Te_all,
+            Wee_all,
+            frac_pos_active_conn,
+        )
 
 
 class Trainer_(Sorn):
@@ -957,30 +970,30 @@ class Trainer_(Sorn):
         """Train the network with the fresh or pretrained network matrices and external stimuli
 
             Args:
-            inputs (np.array, optional): External stimuli. Defaults to None.
+            inputs(np.array, optional): External stimuli. Defaults to None.
 
-            phase (str, optional): Training phase. Defaults to "training".
+            phase(str, optional): Training phase. Defaults to "training".
 
-            matrices (dict, optional): Network states, connections and threshold matrices. Defaults to None.
+            matrices(dict, optional): Network states, connections and threshold matrices. Defaults to None.
 
-            time_steps (int, optional): Total number of time steps to simulate the network. Defaults to 1.
+            time_steps(int, optional): Total number of time steps to simulate the network. Defaults to 1.
 
-            noise (bool, optional): If True, noise will be added. Defaults to True.
+            noise(bool, optional): If True, noise will be added. Defaults to True.
 
-            freeze (list, optional): List of synaptic plasticity mechanisms which will be turned off during simulation. Defaults to None.
+            freeze(list, optional): List of synaptic plasticity mechanisms which will be turned off during simulation. Defaults to None.
 
-            max_workers (int, optional): Maximum workers for multhreading the plasticity steps
+            max_workers(int, optional): Maximum workers for multhreading the plasticity steps
 
         Returns:
-            plastic_matrices (dict): Network states, connections and threshold matrices
+            plastic_matrices(dict): Network states, connections and threshold matrices
 
-            X_all (array): Excitatory network activity collected during entire simulation steps
+            X_all(array): Excitatory network activity collected during entire simulation steps
 
-            Y_all (array): Inhibitory network activity collected during entire simulation steps
+            Y_all(array): Inhibitory network activity collected during entire simulation steps
 
-            R_all (array): Recurrent network activity collected during entire simulation steps
+            R_all(array): Recurrent network activity collected during entire simulation steps
 
-            frac_pos_active_conn (list): Number of positive connection strengths in the network at each time step during simulation"""
+            frac_pos_active_conn(list): Number of positive connection strengths in the network at each time step during simulation"""
 
         assert (
             phase == "plasticity" or "training"
@@ -1009,7 +1022,6 @@ class Trainer_(Sorn):
             if key in kwargs_:
                 setattr(Sorn, key, value)
         Sorn.ni = int(0.2 * Sorn.ne)
-        # assert Sorn.nu == len(inputs[:,0]),"Size mismatch: Input != Nu "
 
         self.phase = phase
         self.matrices = matrices
@@ -1021,6 +1033,8 @@ class Trainer_(Sorn):
         X_all = [0] * self.time_steps
         Y_all = [0] * self.time_steps
         R_all = [0] * self.time_steps
+        Te_all = [0] * self.time_steps
+        Wee_all = [0] * self.time_steps
 
         frac_pos_active_conn = []
 
@@ -1076,10 +1090,33 @@ class Trainer_(Sorn):
             y_buffer[:, 1] = inhibitory_state_yt_buffer.T
 
             if self.phase == "plasticity":
-                Wee[i], Wei[i], Te[i] = Async(max_workers=self.max_workers).step(
-                    x_buffer, y_buffer, Wee[i], Wei[i], Te[i], self.freeze
+                # Plasticity phase
+                plasticity = Plasticity()
+                pool = Pool(processes=4)
+
+                stdp = pool.apply_async(
+                    plasticity.stdp,
+                    [(Wee[i], x_buffer, (0.0, 1.0), "stdp" in self.freeze)],
                 )
 
+                ip = pool.apply_async(
+                    plasticity.ip, [(Te[i], x_buffer, "ip" in self.freeze)]
+                )
+
+                istdp = pool.apply_async(
+                    plasticity.istdp,
+                    [(Wei[i], x_buffer, y_buffer, (0.0, 1.0), "istdp" in self.freeze)],
+                )
+                Wee[i], Te[i], Wei[i] = stdp.get(), ip.get(), istdp.get()
+
+                if "sp" not in self.freeze:
+                    Wee[i] = plasticity.structural_plasticity(Wee[i])
+
+                if "ss" not in self.freeze:
+
+                    Wee[i] = plasticity.ss(Wee[i])
+                    Wei[i] = plasticity.ss(Wei[i])
+
             else:
                 # Wee[i], Wei[i], Te[i] remain same
                 pass
@@ -1092,7 +1129,8 @@ class Trainer_(Sorn):
             X_all[i] = x_buffer[:, 1]
             Y_all[i] = y_buffer[:, 1]
             R_all[i] = r
-
+            Te_all[i] = Te[i]
+            Wee_all[i] = Wee[i]
         plastic_matrices = {
             "Wee": matrix_collection.Wee[-1],
             "Wei": matrix_collection.Wei[-1],
@@ -1103,7 +1141,15 @@ class Trainer_(Sorn):
             "Y": Y[-1],
         }
 
-        return plastic_matrices, X_all, Y_all, R_all, frac_pos_active_conn
+        return (
+            plastic_matrices,
+            X_all,
+            Y_all,
+            R_all,
+            Te_all,
+            Wee_all,
+            frac_pos_active_conn,
+        )
 
 
 Trainer = Trainer_()
